"""
Methods for transforming SICD data to a common state.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from tempfile import mkstemp
import os
from typing import Dict, Tuple, Optional, Union

import numpy
from numpy.polynomial import polynomial
from numpy.random import randn
from scipy.signal import resample

from sarpy.io.general.base import SarpyIOError
from sarpy.processing.ortho_rectify import FullResolutionFetcher
from sarpy.processing.sicd.fft_base import fft, ifft, fftshift, ifftshift, \
    fft_sicd, ifft_sicd

from sarpy.io.complex.base import FlatSICDReader, SICDTypeReader
from sarpy.io.complex.converter import open_complex
from sarpy.io.complex.sicd import SICDWriter
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.Grid import WgtTypeType

logger = logging.getLogger(__name__)


##################
# helper functions

def apply_skew_poly(
        input_data: numpy.ndarray,
        delta_kcoa_poly: numpy.ndarray,
        row_array: numpy.ndarray,
        col_array: numpy.ndarray,
        fft_sgn: int,
        dimension: int,
        forward: bool = False) -> numpy.ndarray:
    """
    Performs the skew operation on the complex array, according to the provided
    delta kcoa polynomial.

    Parameters
    ----------
    input_data : numpy.ndarray
        The input data.
    delta_kcoa_poly : numpy.ndarray
        The delta kcoa polynomial to use.
    row_array : numpy.ndarray
        The row array, should agree with input_data first dimension definition.
    col_array : numpy.ndarray
        The column array, should agree with input_data second dimension definition.
    fft_sgn : int
        The fft sign to use.
    dimension : int
        The dimension to apply along.
    forward : bool
        If True, this shifts forward (i.e. skews), otherwise applies in inverse
        (i.e. deskew) direction.

    Returns
    -------
    numpy.ndarray
    """

    if numpy.all(delta_kcoa_poly == 0):
        return input_data

    delta_kcoa_poly_int = polynomial.polyint(delta_kcoa_poly, axis=dimension)
    if forward:
        fft_sgn *= -1
    return input_data*numpy.exp(1j*fft_sgn*2*numpy.pi*polynomial.polygrid2d(
        row_array, col_array, delta_kcoa_poly_int))


def determine_weight_array(
        input_data_shape: Tuple[int, ...],
        weight_array: numpy.ndarray,
        oversample_rate: Union[int, float],
        dimension: int) -> Tuple[numpy.ndarray, int, int]:
    """
    Determine the appropriate resampled weight array and bounds.

    Parameters
    ----------
    input_data_shape : tuple
        The shape of the input data, which should be a two element tuple.
    weight_array : numpy.ndarray
    oversample_rate : int|float
    dimension : int

    Returns
    -------
    weight_array : numpy.ndarray
        The weight array assuming nominal sampling. In the presence of
        oversampling, this is shorter than relevant dimension of the actual data.
    start_index : int
        The starting index along the given dimension to which to apply weight.
        This will be `0` if not over-sampled.
    end_index : int
        The (non-inclusive) final index along the given dimension to apply
        weight. This will be `input_data_shape[dimension]` if not oversampled.
    """

    if not (isinstance(weight_array, numpy.ndarray) and weight_array.ndim == 1):
        raise ValueError('The weight array must be one-dimensional')

    weight_size = round(input_data_shape[dimension]/oversample_rate)
    if weight_array.ndim != 1:
        raise ValueError('weight_array must be one dimensional.')

    weight_ind_start = int(numpy.floor(0.5*(input_data_shape[dimension] - weight_size)))
    weight_ind_end = weight_ind_start + weight_size

    if weight_array.size == weight_size:
        return weight_array, weight_ind_start, weight_ind_end
    else:
        return resample(weight_array, weight_size), weight_ind_start, weight_ind_end


def apply_weight_array(
        input_data: numpy.ndarray,
        weight_array: numpy.ndarray,
        oversample_rate: Union[int, float],
        dimension: int,
        inverse: bool = False) -> numpy.ndarray:
    """
    Apply the weight array along the given dimension.

    Parameters
    ----------
    input_data : numpy.ndarray
        The complex data array to weight.
    weight_array : numpy.ndarray
        The weight array.
    oversample_rate : int|float
        The oversample rate.
    dimension : int
        Along which dimension to apply the weighting? Must be one of `{0, 1}`.
    inverse : bool
        If `True`, this divides the weight (i.e. de-weight), otherwise it multiplies.

    Returns
    -------
    numpy.ndarray
    """

    if not (isinstance(input_data, numpy.ndarray) and input_data.ndim == 2):
        raise ValueError('The data array must be two-dimensional')

    if weight_array is None:
        # nothing to be done
        return input_data

    weight_array, weight_ind_start, weight_ind_end = determine_weight_array(
        input_data.shape, weight_array, oversample_rate, dimension)

    if inverse and numpy.any(weight_array == 0):
        raise ValueError('inverse=True and the weight array contains some zero entries.')

    output_data = fftshift(fft(input_data, axis=dimension), axes=dimension)
    if dimension == 0:
        if inverse:
            output_data[weight_ind_start:weight_ind_end, :] /= weight_array[:, numpy.newaxis]
        else:
            output_data[weight_ind_start:weight_ind_end, :] *= weight_array[:, numpy.newaxis]
    else:
        if inverse:
            output_data[:, weight_ind_start:weight_ind_end] /= weight_array
        else:
            output_data[:, weight_ind_start:weight_ind_end] *= weight_array
    return ifft(ifftshift(output_data, axes=dimension), axis=dimension)


def _add_poly(poly1: numpy.ndarray, poly2: numpy.ndarray) -> numpy.ndarray:
    """
    Add two-dimensional polynomials together.

    Parameters
    ----------
    poly1 : numpy.ndarray
    poly2 : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(poly1, numpy.ndarray) and poly1.ndim == 2:
        raise TypeError('poly1 must be a two-dimensional numpy array.')
    if not isinstance(poly2, numpy.ndarray) and poly2.ndim == 2:
        raise TypeError('poly2 must be a two-dimensional numpy array.')
    out = numpy.zeros((max(poly1.shape[0], poly2.shape[0]), max(poly1.shape[1], poly2.shape[1])), dtype='float64')
    out[:poly1.shape[0], :poly1.shape[1]] += poly1
    out[:poly2.shape[0], :poly2.shape[1]] += poly2
    return out


def _get_deskew_params(
        the_sicd: SICDType,
        dimension: int) -> Tuple[numpy.ndarray, int]:
    """
    Gets the basic deskew parameters.

    Parameters
    ----------
    the_sicd : SICDType
    dimension : int

    Returns
    -------
    delta_kcoa_poly: numpy.ndarray
    fft_sign : int
    """

    # define the derived variables
    delta_kcoa_poly = numpy.array([[0, ], ], dtype=numpy.float64)
    fft_sign = -1
    if dimension == 0:
        try:
            delta_kcoa_poly = the_sicd.Grid.Row.DeltaKCOAPoly.get_array(dtype='float64')
        except (ValueError, AttributeError):
            pass
        try:
            fft_sign = the_sicd.Grid.Row.Sgn
        except (ValueError, AttributeError):
            pass

    else:
        try:
            delta_kcoa_poly = the_sicd.Grid.Col.DeltaKCOAPoly.get_array(dtype='float64')
        except (ValueError, AttributeError):
            pass
        try:
            fft_sign = the_sicd.Grid.Col.Sgn
        except (ValueError, AttributeError):
            pass
    return delta_kcoa_poly, fft_sign


##########
# sicd state checking functions

def is_not_skewed(sicd: SICDType, dimension: int) -> bool:
    """
    Check if the sicd structure is not skewed along the provided dimension.

    Parameters
    ----------
    sicd : SICDType
    dimension : int

    Returns
    -------
    bool
    """

    if dimension == 0:
        if sicd.Grid is None or sicd.Grid.Row is None or sicd.Grid.Row.DeltaKCOAPoly is None:
            return True
        return numpy.all(sicd.Grid.Row.DeltaKCOAPoly.get_array(dtype='float64') == 0)
    else:
        if sicd.Grid is None or sicd.Grid.Col is None or sicd.Grid.Col.DeltaKCOAPoly is None:
            return True
        return numpy.all(sicd.Grid.Col.DeltaKCOAPoly.get_array(dtype='float64') == 0)


def is_uniform_weight(sicd: SICDType, dimension: int) -> bool:
    """
    Check if the sicd structure is has uniform weight along the provided dimension.

    Parameters
    ----------
    sicd : SICDType
    dimension : int

    Returns
    -------
    bool
    """

    if dimension == 0:
        if sicd.Grid is None or sicd.Grid.Row is None:
            return False
        dir_param = sicd.Grid.Row
    else:
        if sicd.Grid is None or sicd.Grid.Col is None:
            return False
        dir_param = sicd.Grid.Col

    if dir_param.WgtType is not None and dir_param.WgtType.WindowName == 'UNIFORM':
        return True
    if dir_param.WgtFunct is not None:
        return numpy.all(dir_param.WgtFunct == dir_param.WgtFunct[0])
    return True


def is_normalized(sicd: SICDType, dimension: int = 1) -> bool:
    """
    Check if the sicd structure is normalized along the provided dimension.

    Parameters
    ----------
    sicd : SICDType
        The SICD structure.
    dimension : int
        The dimension to test.

    Returns
    -------
    bool
        normalization state in the given dimension
    """

    def _is_fft_sgn_negative() -> bool:
        if dimension == 0:
            if sicd.Grid is None or sicd.Grid.Row is None or sicd.Grid.Row.Sgn is None:
                return True
            return sicd.Grid.Row.Sgn == -1
        else:
            if sicd.Grid is None or sicd.Grid.Col is None or sicd.Grid.Col.Sgn is None:
                return True
            return sicd.Grid.Col.Sgn == -1

    dimension = int(dimension)
    if dimension not in [0, 1]:
        raise ValueError('dimension must be either 0 or 1, got {}'.format(dimension))

    return is_not_skewed(sicd, dimension) and is_uniform_weight(sicd, dimension) and \
        _is_fft_sgn_negative()


###########
# calculator class, intended mainly for use in aperture tool

class DeskewCalculator(FullResolutionFetcher):
    """
    This is a calculator for deskewing/deweighting which requires full resolution
    in both dimensions.
    """

    __slots__ = (
        '_apply_deskew', '_apply_deweighting', '_apply_off_axis', '_delta_kcoa_poly_axis', '_delta_kcoa_poly_off_axis',
        '_row_fft_sgn', '_col_fft_sgn',
        '_row_shift', '_row_mult', '_col_shift', '_col_mult',
        '_row_weight', '_row_pad', '_col_weight', '_col_pad',
        '_is_normalized', '_is_not_skewed_row', '_is_not_skewed_col',
        '_is_uniform_weight_row', '_is_uniform_weight_col', )

    def __init__(self,
                 reader: SICDTypeReader,
                 dimension: int = 1,
                 index: int = 0,
                 apply_deskew: bool = True,
                 apply_deweighting: bool = False,
                 apply_off_axis: bool = True):
        """

        Parameters
        ----------
        reader : SICDTypeReader
        dimension : int
            The dimension in `{0, 1}` along which to deskew. `0` is row/range/fast-time,
            and `1` is column/azimuth/slow-time.
        index : int
            The reader index to utilize
        apply_deskew : bool
            Deskew along the given axis?
        apply_deweighting : bool
            Deweight?
        apply_off_axis : bool
            Deskew off axis, to the extent possible?
        """

        self._apply_deskew = apply_deskew
        self._apply_deweighting = apply_deweighting
        self._apply_off_axis = apply_off_axis
        self._delta_kcoa_poly_axis = None
        self._delta_kcoa_poly_off_axis = None
        self._row_fft_sgn = None
        self._row_shift = None
        self._row_mult = None
        self._col_fft_sgn = None
        self._col_shift = None
        self._col_mult = None
        self._row_weight = None
        self._row_pad = None
        self._col_weight = None
        self._col_pad = None
        self._is_normalized = None
        self._is_not_skewed_row = None
        self._is_not_skewed_col = None
        self._is_uniform_weight_row = None
        self._is_uniform_weight_col = None
        super(DeskewCalculator, self).__init__(
            reader, dimension=dimension, index=index, block_size=None)

    @property
    def dimension(self) -> int:
        """
        int: The dimension along which to perform the color subaperture split.
        """

        return self._dimension

    @dimension.setter
    def dimension(self, value) -> None:
        value = int(value)
        if value not in [0, 1]:
            raise ValueError('dimension must be 0 or 1, got {}'.format(value))
        self._dimension = value
        if self._sicd is not None:
            self._set_sicd(self._sicd)

    def _set_index(self, value) -> None:
        value = int(value)
        if value < 0:
            raise ValueError('The index must be a non-negative integer, got {}'.format(value))

        # noinspection PyUnresolvedReferences
        sicds = self.reader.get_sicds_as_tuple()
        if value >= len(sicds):
            raise ValueError('The index must be less than the sicd count.')
        self._index = value
        self._set_sicd(sicds[value])
        self._data_size = self.reader.get_data_size_as_tuple()[value]

    def _set_sicd(self, the_sicd: SICDType) -> None:
        if the_sicd is None:
            self._sicd = None
            return

        if not isinstance(the_sicd, SICDType):
            raise TypeError('the_sicd must be an insatnce of SICDType, got type {}'.format(type(the_sicd)))

        self._sicd = the_sicd
        row_delta_kcoa_poly, self._row_fft_sgn = _get_deskew_params(the_sicd, 0)
        col_delta_kcoa_poly, self._col_fft_sgn = _get_deskew_params(the_sicd, 1)
        if self.dimension == 0:
            self._delta_kcoa_poly_axis = row_delta_kcoa_poly
            delta_kcoa_poly_int = polynomial.polyint(row_delta_kcoa_poly, axis=0)
            self._delta_kcoa_poly_off_axis = _add_poly(-polynomial.polyder(delta_kcoa_poly_int, axis=1),
                                                       col_delta_kcoa_poly)
        else:
            self._delta_kcoa_poly_axis = col_delta_kcoa_poly
            delta_kcoa_poly_int = polynomial.polyint(col_delta_kcoa_poly, axis=1)
            self._delta_kcoa_poly_off_axis = _add_poly(-polynomial.polyder(delta_kcoa_poly_int, axis=0),
                                                       row_delta_kcoa_poly)

        self._row_shift = the_sicd.ImageData.SCPPixel.Row - the_sicd.ImageData.FirstRow
        self._row_mult = the_sicd.Grid.Row.SS
        self._col_shift = the_sicd.ImageData.SCPPixel.Col - the_sicd.ImageData.FirstCol
        self._col_mult = the_sicd.Grid.Col.SS
        self._row_pad = max(1., 1./(the_sicd.Grid.Row.SS*the_sicd.Grid.Row.ImpRespBW))
        self._row_weight = the_sicd.Grid.Row.WgtFunct.copy() if the_sicd.Grid.Row.WgtFunct is not None else None
        self._col_pad = max(1., 1./(the_sicd.Grid.Col.SS*the_sicd.Grid.Col.ImpRespBW))
        self._col_weight = the_sicd.Grid.Col.WgtFunct.copy() if the_sicd.Grid.Col.WgtFunct is not None else None
        self._is_normalized = is_normalized(the_sicd, self.dimension)
        self._is_not_skewed_row = is_not_skewed(the_sicd, 0)
        self._is_not_skewed_col = is_not_skewed(the_sicd, 1)
        self._is_uniform_weight_row = is_uniform_weight(the_sicd, 0)
        self._is_uniform_weight_col = is_uniform_weight(the_sicd, 1)

    @property
    def apply_deskew(self) -> bool:
        """
        bool: Apply deskew to calculated value. This is for API completeness.
        """

        return self._apply_deskew

    @apply_deskew.setter
    def apply_deskew(self, value):
        self._apply_deskew = (value is True)

    @property
    def apply_deweighting(self) -> bool:
        """
        bool: Apply deweighting to calculated values.
        """

        return self._apply_deweighting

    @apply_deweighting.setter
    def apply_deweighting(self, value):
        self._apply_deweighting = (value is True)

    def _get_index_arrays(
            self,
            row_range: Tuple[int, Union[int, None]],
            row_step: int,
            col_range: Tuple[int, Union[int, None]],
            col_step: int) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Get index array data for polynomial evaluation.

        Parameters
        ----------
        row_range : tuple
        row_step : int
        col_range : tuple
        col_step : int

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
        """

        row_array = self._row_mult*(numpy.arange(row_range[0], -1 if row_range[1] is None else row_range[1], row_step) - self._row_shift)
        col_array = self._col_mult*(numpy.arange(col_range[0], -1 if col_range[1] is None else col_range[1], col_step) - self._col_shift)
        return row_array, col_array

    def __getitem__(self, item) -> numpy.ndarray:
        """
        Fetches the processed data based on the input slice.

        Parameters
        ----------
        item

        Returns
        -------
        numpy.ndarray
        """

        def on_axis_deskew(t_full_data, fft_sgn):
            return apply_skew_poly(
                t_full_data, self._delta_kcoa_poly_axis, row_array, col_array,
                fft_sgn, self.dimension, forward=False)

        def other_axis_deskew(t_full_data, fft_sgn):
            # We cannot generally deskew in both directions at once, but we
            # can recenter the nonskewed dimension with a uniform shift
            if numpy.any(self._delta_kcoa_poly_off_axis != 0):
                # get deltakcoa at midpoint, and treat as a constant polynomial
                row_mid = row_array[int(round(0.5 * row_array.size)) - 1]
                col_mid = col_array[int(round(0.5 * col_array.size)) - 1]
                delta_kcoa_new_const = numpy.zeros((1, 1), dtype='float64')
                delta_kcoa_new_const[0, 0] = polynomial.polyval2d(
                    row_mid, col_mid, self._delta_kcoa_poly_off_axis)
                # apply this uniform shift
                t_full_data = apply_skew_poly(
                    t_full_data, delta_kcoa_new_const, row_array, col_array,
                    fft_sgn, 1-self.dimension, forward=False)
            return t_full_data

        if self._is_normalized or not self.apply_deskew:
            # just fetch the data and return
            if not isinstance(item, tuple) or len(item) != 2:
                raise KeyError(
                    'Slicing in the deskew calculator must be two dimensional. '
                    'Got slice item {}'.format(item))
            return self.reader.__getitem__((item[0], item[1], self.index))

        # parse the slicing to ensure consistent structure
        row_range, col_range, _ = self._parse_slicing(item)
        # get full resolution data in both directions
        row_step = 1 if row_range.step > 0 else -1
        col_step = 1 if col_range.step > 0 else -1
        full_data = self.reader[
               row_range.start:row_range.stop:row_step,
               col_range.start:col_range.stop:col_step,
               self.index]
        # de-weight in each applicable direction
        if self._apply_deweighting and self._is_not_skewed_row and not self._is_uniform_weight_row:
            full_data = apply_weight_array(full_data, self._row_weight, self._row_pad, 0, inverse=True)
        if self._apply_deweighting and self._is_not_skewed_col and not self._is_uniform_weight_col:
            full_data = apply_weight_array(full_data, self._col_weight, self._col_pad, 1, inverse=True)

        # deskew in our given dimension
        row_array, col_array = self._get_index_arrays(
            (row_range.start, row_range.stop), row_step,
            (col_range.start, col_range.stop), col_step)
        if self.dimension == 0:
            # deskew on axis, if necessary
            if not self._is_not_skewed_row:
                full_data = on_axis_deskew(full_data, self._row_fft_sgn)
                if self._apply_deweighting:
                    full_data = apply_weight_array(full_data, self._row_weight, self._row_pad, 0, inverse=True)
            if self._apply_off_axis:
                # deskew off axis, to the extent possible
                full_data = other_axis_deskew(full_data, self._col_fft_sgn)
        elif self.dimension == 1:
            # deskew on axis, if necessary
            if not self._is_not_skewed_col:
                full_data = on_axis_deskew(full_data, self._col_fft_sgn)
                if self._apply_deweighting:
                    full_data = apply_weight_array(full_data, self._col_weight, self._col_pad, 1, inverse=True)
            if self._apply_off_axis:
                # deskew off axis, to the extent possible
                full_data = other_axis_deskew(full_data, self._row_fft_sgn)
        return full_data[::abs(row_range.step), ::abs(col_range.step)]


def aperture_dimension_limits(
        sicd: SICDType,
        dimension: int,
        dimension_limits: Optional[Tuple[Union[int, float], Union[int, float]]] = None,
        aperture_limits: Optional[Tuple[Union[int, float], Union[int, float]]] = None) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """
    This is a helper method to determine the "correct" effective limits for aperture
    processing along the given dimension, considering the ImpRespBW values.

    Parameters
    ----------
    sicd : SICDType
    dimension : int
        One of `{0, 1}`, for the processing dimension
    dimension_limits : None|Tuple[int|float, int|float]
        The base limits along the given dimension, will default to `(0, rows)` for `dimension=0`
        or `(0, columns)` for `dimension=1`, if not provided.
    aperture_limits : None|Tuple[int|float, int|float]
        The desired aperture limits, relative to `dimension_limits`.

    Returns
    -------
    dimension_limits : Tuple[int, int]
        The explicitly populated effective limits along the given dimension.
    aperture_limits : Tuple[int, int]
        The valid aperture limits, relative to `dimension_limits`, after considerations
        of the impulse response bandwidth along the dimension.
    """

    def validate_tuple(tup: Optional[tuple], limit: int) -> Tuple[int, int]:
        if tup is None:
            return 0, limit

        out = int(numpy.floor(tup[0])), int(numpy.ceil(tup[1]))
        if not (0 <= out[0] < out[1] <= limit):
            raise ValueError('Got invalid tuple `{}` for limit `{}`'.format(tup, limit))
        return out

    def extrema_tuple(tup1: tuple, tup2: tuple) -> Tuple[int, int]:
        return int(numpy.floor(max(tup1[0], tup2[0]))), int(numpy.ceil(min(tup1[1], tup2[1])))

    dimension = int(dimension)
    if dimension not in [0, 1]:
        raise ValueError('Got invalid dimension value')

    if dimension == 0:
        the_limit = sicd.ImageData.NumRows
        the_oversample = sicd.Grid.Row.get_oversample_rate()
    else:
        the_limit = sicd.ImageData.NumCols
        the_oversample = sicd.Grid.Col.get_oversample_rate()

    dimension_limits = validate_tuple(dimension_limits, the_limit)
    dimension_count = dimension_limits[1] - dimension_limits[0]
    aperture_limits = validate_tuple(aperture_limits, dimension_count)

    ap_size = dimension_count/the_oversample
    ap_limits = (0.5 * (dimension_count - ap_size), 0.5 * (dimension_count + ap_size))
    ap_limits = extrema_tuple(aperture_limits, ap_limits)
    return dimension_limits, ap_limits


def aperture_dimension_params(
        sicd: SICDType,
        dimension: int,
        dimension_limits: Optional[Tuple[Union[int, float], Union[int, float]]] = None,
        aperture_limits: Optional[Tuple[int, int]] = None,
        new_weight_function: Optional[numpy.ndarray] = None):
    """
    Gets the aperture processing parameters along the given dimension.

    Parameters
    ----------
    sicd : SICDType
    dimension : int
        One of `{0, 1}`, for the processing dimension
    dimension_limits : None|tuple[int|float, int|float]
        The base limits along the given dimension, will default to `(0, rows)`
        for `dimension=0` or `(0, columns)` for `dimension=1`, if not provided.
    aperture_limits : tuple[int, int]
        The valid aperture limits, relative to `dimension_limits`, after
        considerations of the impulse response bandwidth along the dimension.
    new_weight_function : None|numpy.ndarray
        The new weight function. This will default to the current weight function
        if not provided.

    Returns
    -------
    dimension_limits : tuple[int, int]
        The explicitly populated effective limits along the given dimension.
    cur_aperture_limits : tuple[int, int]
        The current valid aperture limits, relative to `dimension_limits`, after
        considerations of the impulse response bandwidth along the dimension.
    cur_aperture_weighting : numpy.ndarray
    new_aperture_limits : tuple[int, int]
        The new valid aperture limits, relative to `dimension_limits`, after
        considerations of the impulse response bandwidth along the dimension.
    new_aperture_weighting : numpy.ndarray
    """

    dimension = int(dimension)
    if dimension not in [0, 1]:
        raise ValueError('Got invalid dimension value')

    dimension_limits, cur_aperture_limits = aperture_dimension_limits(sicd, dimension, dimension_limits, None)
    _, new_aperture_limits = aperture_dimension_limits(sicd, dimension, dimension_limits, aperture_limits)

    cur_ap_count = cur_aperture_limits[1] - cur_aperture_limits[0]
    new_ap_count = new_aperture_limits[1] - new_aperture_limits[0]
    if dimension == 0:
        cur_weight_function = resample(sicd.Grid.Row.WgtFunct, cur_ap_count)
        if new_weight_function is None:
            new_weight_function = resample(sicd.Grid.Row.WgtFunct, new_ap_count)
        else:
            new_weight_function = resample(new_weight_function, new_ap_count)
    else:
        cur_weight_function = resample(sicd.Grid.Col.WgtFunct, cur_ap_count)
        if new_weight_function is None:
            new_weight_function = resample(sicd.Grid.Col.WgtFunct, new_ap_count)
        else:
            new_weight_function = resample(new_weight_function, new_ap_count)
    return dimension_limits, cur_aperture_limits, cur_weight_function, new_aperture_limits, new_weight_function


def noise_scaling(
        cur_ap_limits: Tuple[int, int],
        cur_weighting: numpy.ndarray,
        new_ap_limits: Tuple[int, int],
        new_weighting: numpy.ndarray) -> float:
    """
    Gets noise scaling due to sub-aperture degradation and re-weighting along one
    dimension.

    Parameters
    ----------
    cur_ap_limits : Tuple[int, int]
    cur_weighting : numpy.ndarray
    new_ap_limits : Tuple[int, int]
    new_weighting : numpy.ndarray

    Returns
    -------
    noise_multiplier : float
    """

    start_lim = new_ap_limits[0]-cur_ap_limits[0]
    end_lim = start_lim + new_ap_limits[1] - new_ap_limits[0]
    weight_change = new_weighting/cur_weighting[start_lim:end_lim]
    second_moment = numpy.sum(weight_change**2)/float(cur_weighting.size)
    return second_moment


def sicd_degrade_reweight(
        reader: SICDTypeReader,
        output_file: Optional[str] = None,
        index: int = 0,
        row_limits: Optional[Tuple[int, int]] = None,
        column_limits: Optional[Tuple[int, int]] = None,
        row_aperture: Optional[Tuple[int, int]] = None,
        row_weighting: Optional[Dict] = None,
        column_aperture: Optional[Tuple[int, int]] = None,
        column_weighting: Optional[Dict] = None,
        add_noise: Optional[float] = None,
        pixel_threshold: Optional[int] = 1500*1500,
        check_existence: bool = True,
        check_older_version: bool = False,
        repopulate_rniirs: bool = True) -> Optional[FlatSICDReader]:
    r"""
    Given input, create a SICD (file or reader) with modified weighting/subaperture
    parameters. Any additional noise will be added **before** performing any sub-aperture
    degradation processing.

    Recall that reducing the size of the impulse response bandwidth via sub-aperture
    degradation in a single direction by by :math:`ratio`, actually decreases the
    magnitude of the noise in pixel power by :math:`ratio`, or subtracts
    :math:`10*\log_{10}(ratio)` from the noise given in dB.

    .. warning::

        To ensure correctness of metadata, if the Noise Polynomial is present,
        then then the NoiseLevelType must be `'ABSOLUTE'`. Otherwise an exception
        will be raised.

    Parameters
    ----------
    reader : str|SICDTypeReader
        A sicd type reader.
    output_file : None|str
        If `None`, an in-memory SICD reader instance will be returned. Otherwise,
        this is the path for the produced output SICD file.
    index : int
        The reader index to be used.
    row_limits : None|(int, int)
        Row limits for the underlying data.
    column_limits : None|(int, int)
        Column limits for the underlying data.
    row_aperture : None|tuple
        `None` (no row subaperture), or integer valued `start_row, end_row` for the
        row subaperture definition. This is with respect to row values AFTER
        considering `row_limits`. Note that this will reduce the noise, so the
        noise polynomial (if it is populated) will be modified.
    row_weighting : None|dict
        `None` (no row weighting change), or the new row weighting parameters
        `{'WindowName': <name>, 'Parameters: {}, 'WgtFunct': array}`.
    column_aperture : None|tuple
        `None` (no column subaperture), or integer valued `start_col, end_col` for the
        column sub-aperture definition. This is with respect to row values AFTER
        considering `column_limits`. Note that this will reduce the noise, so the
        noise polynomial (if it is populated) will be modified.
    column_weighting : None|dict
        `None` (no columng weighting change), or the new column weighting parameters
        `{'WindowName': <name>, 'Parameters: {}, 'WgtFunct': array}`.
    add_noise : None|float
        If provided, Gaussian white noise of pixel power `add_noise` will be
        added, prior to any subbaperture processing. Note that performing subaperture
        processing actually reduces the resulting noise, which will also be considered.
    pixel_threshold : None|int
        Approximate pixel area threshold for performing this directly in memory.
    check_existence : bool
        Should we check if the given file already exists, and raise an exception if so?
    check_older_version : bool
        Try to use a less recent version of SICD (1.1), for possible application compliance issues?
    repopulate_rniirs : bool
        Should we try to repopulate the estimated RNIIRS value?

    Returns
    -------
    None|FlatSICDReader
        No return if `output_file` is provided, otherwise the returns the in-memory
        reader object.
    """

    def validate_filename():
        if output_file is None:
            return

        if check_existence and os.path.exists(output_file):
            raise SarpyIOError('The file {} already exists.'.format(output_file))

    def validate_sicd(the_sicd):
        if the_sicd.Grid is None or the_sicd.Grid.Row is None or the_sicd.Grid.Col is None:
            raise ValueError('Grid.Row and Grid.Col must be populated')
        for direction in ['Row', 'Col']:
            el = getattr(the_sicd.Grid, direction)
            if el.DeltaKCOAPoly is None or el.WgtFunct is None:
                raise ValueError('DeltaKCOAPoly and WgtFunct must be populated for both Row and Col')

    def validate_limits(lims, max_index):
        if lims is None:
            return 0, max_index
        _lims = (int(lims[0]), int(lims[1]))
        if not (0 <= _lims[0] < _lims[1] <= max_index):
            raise ValueError('Got poorly formatted index limit {}'.format(lims))
        return _lims

    def get_iterations(max_index, other_index):
        if in_memory:
            return [(0, max_index), ]
        out = []
        block = int(pixel_threshold / float(other_index))
        _start_ind = 0
        while _start_ind < max_index:
            _end_ind = min(_start_ind + block, max_index)
            out.append((_start_ind, _end_ind))
            _start_ind = _end_ind
        return out

    def get_direction_array_meters(dimension, start_index, end_index):
        if dimension == 0:
            shift = sicd.ImageData.FirstRow - sicd.ImageData.SCPPixel.Row
            multiplier = sicd.Grid.Row.SS
        else:
            shift = sicd.ImageData.FirstCol - sicd.ImageData.SCPPixel.Col
            multiplier = sicd.Grid.Col.SS
        return (numpy.arange(start_index, end_index) + shift)*multiplier

    def do_add_noise():
        if add_noise is None:
            return

        # noinspection PyBroadException
        try:
            variance = float(add_noise)
            if variance <= 0:
                logger.error('add_noise was provided as `{}`, but must be a positive number'.format(add_noise))
                return
        except Exception:
            logger.error('add_noise was provided as `{}`, but must be a positive number'.format(add_noise))
            return

        sigma = numpy.sqrt(0.5*variance)
        if noise_level is not None:
            noise_constant_power = numpy.exp(numpy.log(10)*0.1*noise_level.NoisePoly[0, 0])
            noise_constant_power += variance
            noise_constant_db = 10*numpy.log10(noise_constant_power)
            noise_level.NoisePoly[0, 0] = noise_constant_db

        for (_start_ind, _stop_ind) in row_iterations:
            d_shape = (_stop_ind - _start_ind, data_shape[1])
            added_noise = numpy.empty(d_shape, dtype='complex64')
            added_noise[:].real = randn(*d_shape).astype('float32')
            added_noise[:].imag = randn(*d_shape).astype('float32')
            added_noise *= sigma
            working_data[_start_ind:_stop_ind, :] += added_noise

    def do_dimension(dimension):
        if dimension == 0:
            dir_params = sicd.Grid.Row
            aperture_in = row_aperture
            weighting_in = row_weighting
            dimension_limits = row_limits
        else:
            dir_params = sicd.Grid.Col
            aperture_in = column_aperture
            weighting_in = column_weighting
            dimension_limits = column_limits

        not_skewed = is_not_skewed(sicd, dimension)
        uniform_weight = is_uniform_weight(sicd, dimension)
        delta_kcoa = dir_params.DeltaKCOAPoly.get_array(dtype='float64')

        st_beam_comp = sicd.ImageFormation.STBeamComp if sicd.ImageFormation is not None else None

        if dimension == 1 and (not uniform_weight or weighting_in is not None):
            if st_beam_comp is None:
                logger.warning(
                    'Processing along the column direction requires modification\n\t'
                    'of the original weighting scheme, and the value for\n\t'
                    'ImageFormation.STBeamComp is not populated.\n\t'
                    'It is unclear how imperfect the de-weighting effort along the column may be.')
            elif st_beam_comp == 'NO':
                logger.warning(
                    'Processing along the column direction requires modification\n\t'
                    'of the original weighting scheme, and the value for\n\t'
                    'ImageFormation.STBeamComp is populated as `NO`.\n\t'
                    'It is likely that the de-weighting effort along the column is imperfect.')

        if aperture_in is None and weighting_in is None:
            # nothing to be done in this dimension
            return noise_adjustment_multiplier

        new_weight = None if weighting_in is None else weighting_in['WgtFunct']
        dimension_limits, cur_aperture_limits, cur_weight_function, \
            new_aperture_limits, new_weight_function = aperture_dimension_params(
                old_sicd, dimension, dimension_limits=dimension_limits,
                aperture_limits=aperture_in, new_weight_function=new_weight)
        index_count = dimension_limits[1] - dimension_limits[0]
        cur_center_index = 0.5*(cur_aperture_limits[0] + cur_aperture_limits[1])
        new_center_index = 0.5*(new_aperture_limits[0] + new_aperture_limits[1])
        noise_multiplier = noise_scaling(
            cur_aperture_limits, cur_weight_function, new_aperture_limits, new_weight_function)

        # perform deskew, if necessary
        if not not_skewed:
            if dimension == 0:
                row_array = get_direction_array_meters(0, 0, out_data_shape[0])
                for (_start_ind, _stop_ind) in col_iterations:
                    col_array = get_direction_array_meters(1, _start_ind, _stop_ind)
                    working_data[:, _start_ind:_stop_ind] = apply_skew_poly(
                        working_data[:, _start_ind:_stop_ind], delta_kcoa,
                        row_array, col_array, dir_params.Sgn, 0, forward=False)
            else:
                col_array = get_direction_array_meters(1, 0, out_data_shape[1])
                for (_start_ind, _stop_ind) in row_iterations:
                    row_array = get_direction_array_meters(0, _start_ind, _stop_ind)
                    working_data[_start_ind:_stop_ind, :] = apply_skew_poly(
                        working_data[_start_ind:_stop_ind, :], delta_kcoa,
                        row_array, col_array, dir_params.Sgn, 1, forward=False)

        # perform fourier transform along the given dimension
        if dimension == 0:
            for (_start_ind, _stop_ind) in col_iterations:
                working_data[:, _start_ind:_stop_ind] = fftshift(
                    fft_sicd(working_data[:, _start_ind:_stop_ind], dimension, sicd), axes=dimension)
        else:
            for (_start_ind, _stop_ind) in row_iterations:
                working_data[_start_ind:_stop_ind, :] = fftshift(
                    fft_sicd(working_data[_start_ind:_stop_ind, :], dimension, sicd), axes=dimension)

        # perform deweight, if necessary
        if not uniform_weight:
            if dimension == 0:
                working_data[cur_aperture_limits[0]:cur_aperture_limits[1], :] /= cur_weight_function[:, numpy.newaxis]
            else:
                working_data[:, cur_aperture_limits[0]:cur_aperture_limits[1]] /= cur_weight_function

        # do sub-aperture, if necessary
        if aperture_in is not None:
            if dimension == 0:
                working_data[:new_aperture_limits[0], :] = 0
                working_data[new_aperture_limits[1]:, :] = 0
            else:
                working_data[:, :new_aperture_limits[0]] = 0
                working_data[:, new_aperture_limits[1]:] = 0

            the_ratio = float(new_aperture_limits[1] - new_aperture_limits[0]) / \
                float(cur_aperture_limits[1] - cur_aperture_limits[0])
            # modify the ImpRespBW value (derived ImpRespWid handled at the end)
            dir_params.ImpRespBW *= the_ratio

        # perform reweight, if necessary
        if weighting_in is not None:
            if dimension == 0:
                working_data[new_aperture_limits[0]:new_aperture_limits[1], :] *= new_weight_function[:, numpy.newaxis]
            else:
                working_data[:, new_aperture_limits[0]:new_aperture_limits[1]] *= new_weight_function
            # modify the weight definition
            dir_params.WgtType = WgtTypeType(
                WindowName=weighting_in['WindowName'],
                Parameters=weighting_in.get('Parameters', None))
            dir_params.WgtFunct = weighting_in['WgtFunct'].copy()
        elif not uniform_weight:
            if dimension == 0:
                working_data[new_aperture_limits[0]:new_aperture_limits[1], :] *= new_weight_function[:, numpy.newaxis]
            else:
                working_data[:, new_aperture_limits[0]:new_aperture_limits[1]] *= new_weight_function

        # perform inverse fourier transform along the given dimension
        if dimension == 0:
            for (_start_ind, _stop_ind) in col_iterations:
                working_data[:, _start_ind:_stop_ind] = ifft_sicd(
                    ifftshift(working_data[:, _start_ind:_stop_ind], axes=dimension), dimension, sicd)
        else:
            for (_start_ind, _stop_ind) in row_iterations:
                working_data[_start_ind:_stop_ind, :] = ifft_sicd(
                    ifftshift(working_data[_start_ind:_stop_ind, :], axes=dimension), dimension, sicd)

        # perform the (original) reskew, if necessary
        if not numpy.all(delta_kcoa == 0):
            if dimension == 0:
                row_array = get_direction_array_meters(0, 0, out_data_shape[0])
                for (_start_ind, _stop_ind) in col_iterations:
                    col_array = get_direction_array_meters(1, _start_ind, _stop_ind)
                    working_data[:, _start_ind:_stop_ind] = apply_skew_poly(
                        working_data[:, _start_ind:_stop_ind], delta_kcoa,
                        row_array, col_array, dir_params.Sgn, 0, forward=True)
            else:
                col_array = get_direction_array_meters(1, 0, out_data_shape[1])
                for (_start_ind, _stop_ind) in row_iterations:
                    row_array = get_direction_array_meters(0, _start_ind, _stop_ind)
                    working_data[_start_ind:_stop_ind, :] = apply_skew_poly(
                        working_data[_start_ind:_stop_ind, :], delta_kcoa,
                        row_array, col_array, dir_params.Sgn, 1, forward=True)

        # modify the delta_kcoa_poly - introduce the shift necessary for additional offset
        if new_center_index != cur_center_index:
            additional_shift = dir_params.Sgn*(cur_center_index - new_center_index)/float(index_count*dir_params.SS)
            delta_kcoa[0, 0] += additional_shift
            dir_params.DeltaKCOAPoly = delta_kcoa

        return noise_adjustment_multiplier*noise_multiplier

    if isinstance(reader, str):
        reader = open_complex(reader)

    if not isinstance(reader, SICDTypeReader):
        raise TypeError('reader must be sicd type reader, got {}'.format(reader))

    # noinspection PyUnresolvedReferences
    old_sicd = reader.get_sicds_as_tuple()[index]
    validate_sicd(old_sicd)
    validate_filename()

    data_shape = reader.get_data_size_as_tuple()[index]

    redo_geo = (row_limits is not None or column_limits is not None)

    row_limits = validate_limits(row_limits, data_shape[0])
    column_limits = validate_limits(column_limits, data_shape[1])

    # prepare our working sicd structure
    sicd = old_sicd.copy()  # type: SICDType

    noise_level = None if sicd.Radiometric is None else sicd.Radiometric.NoiseLevel
    # NB: as an alternative, we could drop the noise polynomial?
    if add_noise is not None and \
            noise_level is not None and \
            noise_level.NoiseLevelType != 'ABSOLUTE':
        raise ValueError(
            'add_noise is provided, but the radiometric noise is populated,\n\t'
            'with `NoiseLevelType={}`'.format(noise_level.NoiseLevelType))

    sicd.ImageData.FirstRow += row_limits[0]
    sicd.ImageData.NumRows = row_limits[1] - row_limits[0]
    sicd.ImageData.FirstCol += column_limits[0]
    sicd.ImageData.NumCols = column_limits[1] - column_limits[0]
    if redo_geo:
        sicd.define_geo_image_corners(override=True)

    out_data_shape = (sicd.ImageData.NumRows, sicd.ImageData.NumCols)

    pixel_area = out_data_shape[0]*out_data_shape[1]
    temp_file = None
    in_memory = True if (output_file is None or pixel_threshold is None) else (pixel_area < pixel_threshold)

    row_iterations = get_iterations(out_data_shape[0], out_data_shape[1])
    col_iterations = get_iterations(out_data_shape[1], out_data_shape[0])

    if in_memory:
        working_data = reader[row_limits[0]:row_limits[1], column_limits[0]:column_limits[1], index]
    else:
        _, temp_file = mkstemp(suffix='.sarpy.cache', text=False)
        working_data = numpy.memmap(temp_file, dtype='complex64', mode='r+', offset=0, shape=out_data_shape)
        for (start_ind, stop_ind) in row_iterations:
            working_data[start_ind:stop_ind, :] = reader[
                                                  start_ind+row_limits[0]:stop_ind+row_limits[0],
                                                  column_limits[0]:column_limits[1], index]

    noise_adjustment_multiplier = 1.0

    # NB: I'm adding Gaussian white noise first
    do_add_noise()

    # do, as necessary, along the row
    noise_adjustment_multiplier = do_dimension(0)
    # do, as necessary, along the row
    noise_adjustment_multiplier = do_dimension(1)

    # re-derive the various ImpResp parameters
    sicd.Grid.derive_direction_params(sicd.ImageData, populate=True)

    # adjust the noise polynomial
    if noise_level is not None:
        noise_level.NoisePoly[0, 0] += 10*numpy.log10(noise_adjustment_multiplier)

    if sicd.Radiometric is not None and sicd.Radiometric.RCSSFPoly is not None:
        sf_adjust = old_sicd.Grid.get_slant_plane_area()/noise_adjustment_multiplier
        sicd.Radiometric.RCSSFPoly.Coefs = sicd.Radiometric.RCSSFPoly.get_array()*sf_adjust
        sicd.Radiometric.BetaZeroSFPoly.Coefs = sicd.Radiometric.BetaZeroSFPoly.get_array() / \
            noise_adjustment_multiplier
        sicd.Radiometric.SigmaZeroSFPoly.Coefs = sicd.Radiometric.SigmaZeroSFPoly.get_array() / \
            noise_adjustment_multiplier
        sicd.Radiometric.GammaZeroSFPoly.Coefs = sicd.Radiometric.GammaZeroSFPoly.get_array() / \
            noise_adjustment_multiplier

    if sicd.RMA is not None and sicd.RMA.INCA is not None and sicd.RMA.INCA.TimeCAPoly is not None:
        # redefine the INCA doppler centroid poly to be in keeping with any redefinition of our Col.DeltaKCOAPoly?
        sicd.RMA.INCA.DopCentroidPoly = sicd.Grid.Col.DeltaKCOAPoly.get_array(dtype='float64') / \
            sicd.RMA.INCA.TimeCAPoly[1]

    if repopulate_rniirs:
        sicd.populate_rniirs(override=True)

    if output_file is None:
        return FlatSICDReader(sicd, working_data)
    else:
        # write out the new sicd file
        with SICDWriter(
                output_file, sicd,
                check_older_version=check_older_version, check_existence=check_existence) as writer:
            for (start_ind, stop_ind) in row_iterations:
                writer.write_chip(working_data[start_ind:stop_ind, :], start_indices=(start_ind, 0))

        if temp_file is not None and os.path.exists(temp_file):
            working_data = None
            os.remove(temp_file)



"""
The SICDType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from copy import deepcopy
import re
from collections import OrderedDict
from typing import Optional, Dict, Union, Tuple

import numpy

from sarpy.geometry import point_projection
from sarpy.io.complex.naming.utils import get_sicd_name
from sarpy.io.complex.sicd_schema import get_urn_details, get_default_tuple, \
    get_default_version_string, get_specification_identifier, WRITABLE_VERSIONS

from sarpy.io.xml.base import Serializable, parse_xml_from_file, parse_xml_from_string
from sarpy.io.xml.descriptors import SerializableDescriptor

from .base import DEFAULT_STRICT
from .CollectionInfo import CollectionInfoType
from .ImageCreation import ImageCreationType
from .ImageData import ImageDataType
from .GeoData import GeoDataType
from .GeoData import SCPType
from .Grid import GridType
from .Timeline import TimelineType
from .Position import PositionType
from .RadarCollection import RadarCollectionType
from .ImageFormation import ImageFormationType
from .SCPCOA import SCPCOAType
from .Radiometric import RadiometricType
from .Antenna import AntennaType
from .ErrorStatistics import ErrorStatisticsType
from .MatchInfo import MatchInfoType
from .RgAzComp import RgAzCompType
from .PFA import PFAType
from .RMA import RMAType
from .validation_checks import detailed_validation_checks


logger = logging.getLogger(__name__)

#########
# Module variables
_SICD_SPECIFICATION_IDENTIFIER = get_specification_identifier()
_SICD_DEFAULT_TUPLE = get_default_tuple()
_SICD_VERSION_DEFAULT = get_default_version_string()
_SICD_SPEC_DETAILS = {
    key: {
        'namespace': 'urn:SICD:{}'.format(key),
        'details': get_urn_details(key)}
    for key in WRITABLE_VERSIONS}


class SICDType(Serializable):
    """
    Sensor Independent Complex Data object, containing all the relevant data to formulate products.
    """

    _fields = (
        'CollectionInfo', 'ImageCreation', 'ImageData', 'GeoData', 'Grid', 'Timeline', 'Position',
        'RadarCollection', 'ImageFormation', 'SCPCOA', 'Radiometric', 'Antenna', 'ErrorStatistics',
        'MatchInfo', 'RgAzComp', 'PFA', 'RMA')
    _required = (
        'CollectionInfo', 'ImageData', 'GeoData', 'Grid', 'Timeline', 'Position',
        'RadarCollection', 'ImageFormation', 'SCPCOA')
    _choice = ({'required': False, 'collection': ('RgAzComp', 'PFA', 'RMA')}, )
    # descriptors
    CollectionInfo = SerializableDescriptor(
        'CollectionInfo', CollectionInfoType, _required, strict=False,
        docstring='General information about the '
                  'collection.')  # type: CollectionInfoType
    ImageCreation = SerializableDescriptor(
        'ImageCreation', ImageCreationType, _required, strict=False,
        docstring='General information about the image '
                  'creation.')  # type: Optional[ImageCreationType]
    ImageData = SerializableDescriptor(
        'ImageData', ImageDataType, _required, strict=False,  # it is senseless to not have this element
        docstring='The image pixel data.')  # type: ImageDataType
    GeoData = SerializableDescriptor(
        'GeoData', GeoDataType, _required, strict=False,
        docstring='The geographic coordinates of the image coverage area.')  # type: GeoDataType
    Grid = SerializableDescriptor(
        'Grid', GridType, _required, strict=False,
        docstring='The image sample grid.')  # type: GridType
    Timeline = SerializableDescriptor(
        'Timeline', TimelineType, _required, strict=False,
        docstring='The imaging collection time line.')  # type: TimelineType
    Position = SerializableDescriptor(
        'Position', PositionType, _required, strict=False,
        docstring='The platform and ground reference point coordinates as a function of time.')  # type: PositionType
    RadarCollection = SerializableDescriptor(
        'RadarCollection', RadarCollectionType, _required, strict=False,
        docstring='The radar collection information.')  # type: RadarCollectionType
    ImageFormation = SerializableDescriptor(
        'ImageFormation', ImageFormationType, _required, strict=False,
        docstring='The image formation process.')  # type: ImageFormationType
    SCPCOA = SerializableDescriptor(
        'SCPCOA', SCPCOAType, _required, strict=False,
        docstring='*Center of Aperture (COA)* for the *Scene Center Point (SCP)*.')  # type: SCPCOAType
    Radiometric = SerializableDescriptor(
        'Radiometric', RadiometricType, _required, strict=False,
        docstring='The radiometric calibration '
                  'parameters.')  # type: Optional[RadiometricType]
    Antenna = SerializableDescriptor(
        'Antenna', AntennaType, _required, strict=False,
        docstring='Parameters that describe the antenna illumination patterns '
                  'during the collection.')  # type: Optional[AntennaType]
    ErrorStatistics = SerializableDescriptor(
        'ErrorStatistics', ErrorStatisticsType, _required, strict=False,
        docstring='Parameters used to compute error statistics within the '
                  '*SICD* sensor model.')  # type: Optional[ErrorStatisticsType]
    MatchInfo = SerializableDescriptor(
        'MatchInfo', MatchInfoType, _required, strict=False,
        docstring='Information about other collections that are matched to the '
                  'current collection. The current collection is the collection '
                  'from which this *SICD* product was '
                  'generated.')  # type: Optional[MatchInfoType]
    RgAzComp = SerializableDescriptor(
        'RgAzComp', RgAzCompType, _required, strict=False,
        docstring='Parameters included for a *Range, Doppler* '
                  'image.')  # type: Optional[RgAzCompType]
    PFA = SerializableDescriptor(
        'PFA', PFAType, _required, strict=False,
        docstring='Parameters included when the image is formed using the '
                  '*Polar Formation Algorithm (PFA)*.')  # type: Optional[PFAType]
    RMA = SerializableDescriptor(
        'RMA', RMAType, _required, strict=False,
        docstring='Parameters included when the image is formed using the '
                  '*Range Migration Algorithm (RMA)*.')  # type: Optional[RMAType]

    def __init__(
            self,
            CollectionInfo: CollectionInfoType = None,
            ImageCreation: Optional[ImageCreationType] = None,
            ImageData: ImageDataType = None,
            GeoData: GeoDataType = None,
            Grid: GridType = None,
            Timeline: TimelineType = None,
            Position: PositionType = None,
            RadarCollection: RadarCollectionType = None,
            ImageFormation: ImageFormationType = None,
            SCPCOA: SCPCOAType = None,
            Radiometric: Optional[RadiometricType] = None,
            Antenna: Optional[AntennaType] = None,
            ErrorStatistics: Optional[ErrorStatisticsType] = None,
            MatchInfo: Optional[MatchInfoType] = None,
            RgAzComp: Optional[RgAzCompType] = None,
            PFA: Optional[PFAType] = None,
            RMA: Optional[RMAType] = None,
            **kwargs):
        """

        Parameters
        ----------
        CollectionInfo : CollectionInfoType
        ImageCreation : ImageCreationType
        ImageData : ImageDataType
        GeoData : GeoDataType
        Grid : GridType
        Timeline : TimelineType
        Position : PositionType
        RadarCollection : RadarCollectionType
        ImageFormation : ImageFormationType
        SCPCOA : SCPCOAType
        Radiometric : RadiometricType
        Antenna : AntennaType
        ErrorStatistics : ErrorStatisticsType
        MatchInfo : MatchInfoType
        RgAzComp : RgAzCompType
        PFA : PFAType
        RMA : RMAType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        nitf = kwargs.get('_NITF', {})
        if not isinstance(nitf, dict):
            raise TypeError('Provided NITF options are required to be in dictionary form.')
        self._NITF = nitf

        self._coa_projection = None
        self.CollectionInfo = CollectionInfo
        self.ImageCreation = ImageCreation
        self.ImageData = ImageData
        self.GeoData = GeoData
        self.Grid = Grid
        self.Timeline = Timeline
        self.Position = Position
        self.RadarCollection = RadarCollection
        self.ImageFormation = ImageFormation
        self.SCPCOA = SCPCOA
        self.Radiometric = Radiometric
        self.Antenna = Antenna
        self.ErrorStatistics = ErrorStatistics
        self.MatchInfo = MatchInfo
        self.RgAzComp = RgAzComp
        self.PFA = PFA
        self.RMA = RMA
        super(SICDType, self).__init__(**kwargs)

    @property
    def coa_projection(self):
        """
        The COA Projection object, if previously defined through using :func:`define_coa_projection`.

        Returns
        -------
        None|sarpy.geometry.point_projection.COAProjection
        """

        return self._coa_projection

    @property
    def NITF(self) -> Optional[Dict]:
        """
        Optional dictionary of NITF header information, pertains only to subsequent
        SICD file writing.

        Returns
        -------
        Dict
        """

        return self._NITF

    @NITF.setter
    def NITF(self, value: Optional[Dict]):
        if value is None:
            self._NITF = {}
            return
        if isinstance(value, dict):
            self._NITF = value
        if not isinstance(value, dict):
            raise TypeError('data must be dictionary instance. Received {}'.format(type(value)))

    @property
    def ImageFormType(self) -> str:
        """
        str: *READ ONLY* Identifies the specific image formation type supplied. This is determined by
        returning the (first) attribute among `RgAzComp`, `PFA`, `RMA` which is populated. `OTHER` will be returned if
        none of them are populated.
        """

        for attribute in self._choice[0]['collection']:
            if getattr(self, attribute) is not None:
                return attribute
        return 'OTHER'

    def update_scp(
            self,
            point: Union[numpy.ndarray, list, tuple],
            coord_system: str = 'ECF'):
        """
        Modify the SCP point, and modify the associated SCPCOA fields.

        Parameters
        ----------
        point : numpy.ndarray|tuple|list
        coord_system : str
            Either 'ECF' or 'LLH', and 'ECF' will take precedence.

        Returns
        -------
        None
        """

        if isinstance(point, (list, tuple)):
            point = numpy.array(point, dtype='float64')
        if not isinstance(point, numpy.ndarray):
            raise TypeError('point must be an numpy.ndarray')
        if point.shape != (3, ):
            raise ValueError('point must be a one-dimensional, 3 element array')
        if coord_system == 'LLH':
            self.GeoData.SCP.LLH = point
        else:
            self.GeoData.SCP.ECF = point

        if self.SCPCOA is not None:
            self.SCPCOA.rederive(self.Grid, self.Position, self.GeoData)

    def _basic_validity_check(self) -> bool:
        condition = super(SICDType, self)._basic_validity_check()
        condition &= detailed_validation_checks(self)
        return condition

    def is_valid(self, recursive: bool = False, stack: bool = False) -> bool:
        all_required = self._basic_validity_check()
        if not recursive:
            return all_required

        valid_children = self._recursive_validity_check(stack=stack)
        return all_required & valid_children

    def define_geo_image_corners(self, override: bool = False) -> None:
        """
        Defines the GeoData image corner points (if possible), if they are not already defined.

        Returns
        -------
        None
        """

        if self.GeoData is None:
            try:
                self.GeoData = GeoDataType(SCP=SCPType(self.RadarCollection.Area.Plane.RefPt.ECF))
            except AttributeError:
                return

        if self.GeoData.ImageCorners is not None and not override:
            return  # nothing to be done

        try:
            vertex_data = self.ImageData.get_full_vertex_data(dtype=numpy.float64)
            corner_coords = self.project_image_to_ground_geo(vertex_data)
        except (ValueError, AttributeError):
            return

        self.GeoData.ImageCorners = corner_coords

    def define_geo_valid_data(self) -> None:
        """
        Defines the GeoData valid data corner points (if possible), if they are not already defined.

        Returns
        -------
        None
        """

        if self.GeoData is None or self.GeoData.ValidData is not None:
            return  # nothing to be done

        try:
            valid_vertices = self.ImageData.get_valid_vertex_data(dtype=numpy.float64)
            if valid_vertices is not None:
                self.GeoData.ValidData = self.project_image_to_ground_geo(valid_vertices)
        except AttributeError:
            pass

    def derive(self) -> None:
        """
        Populates any potential derived data in the SICD structure. This should get called after reading an XML,
        or as a user desires.

        Returns
        -------
        None
        """

        # Note that there is dependency in calling order between steps - don't naively rearrange the following.
        if self.SCPCOA is None:
            self.SCPCOA = SCPCOAType()

        # noinspection PyProtectedMember
        self.SCPCOA._derive_scp_time(self.Grid)

        if self.Grid is not None:
            # noinspection PyProtectedMember
            self.Grid._derive_time_coa_poly(self.CollectionInfo, self.SCPCOA)

        # noinspection PyProtectedMember
        self.SCPCOA._derive_position(self.Position)

        if self.Position is None and self.SCPCOA.ARPPos is not None and \
                self.SCPCOA.ARPVel is not None and self.SCPCOA.SCPTime is not None:
            self.Position = PositionType()  # important parameter derived in the next step
        if self.Position is not None:
            # noinspection PyProtectedMember
            self.Position._derive_arp_poly(self.SCPCOA)

        if self.GeoData is not None:
            self.GeoData.derive()  # ensures both coordinate systems are defined for SCP

        if self.Grid is not None:
            # noinspection PyProtectedMember
            self.Grid.derive_direction_params(self.ImageData)

        if self.RadarCollection is not None:
            self.RadarCollection.derive()

        if self.ImageFormation is not None:
            # call after RadarCollection.derive(), and only if the entire transmitted bandwidth was used to process.
            # noinspection PyProtectedMember
            self.ImageFormation._derive_tx_frequency_proc(self.RadarCollection)

        # noinspection PyProtectedMember
        self.SCPCOA._derive_geometry_parameters(self.GeoData)

        # verify ImageFormation things make sense
        im_form_algo = None
        if self.ImageFormation is not None and self.ImageFormation.ImageFormAlgo is not None:
            im_form_algo = self.ImageFormation.ImageFormAlgo.upper()
        if im_form_algo == 'RGAZCOMP':
            # Check Grid settings
            if self.Grid is None:
                self.Grid = GridType()
            # noinspection PyProtectedMember
            self.Grid._derive_rg_az_comp(self.GeoData, self.SCPCOA, self.RadarCollection, self.ImageFormation)

            # Check RgAzComp settings
            if self.RgAzComp is None:
                self.RgAzComp = RgAzCompType()
            # noinspection PyProtectedMember
            self.RgAzComp._derive_parameters(self.Grid, self.Timeline, self.SCPCOA)
        elif im_form_algo == 'PFA':
            if self.PFA is None:
                self.PFA = PFAType()
            # noinspection PyProtectedMember
            self.PFA._derive_parameters(self.Grid, self.SCPCOA, self.GeoData, self.Position, self.Timeline)

            if self.Grid is not None:
                # noinspection PyProtectedMember
                self.Grid._derive_pfa(
                    self.GeoData, self.RadarCollection, self.ImageFormation, self.Position, self.PFA)
        elif im_form_algo == 'RMA' or self.RMA is not None:
            if self.RMA is not None:
                # noinspection PyProtectedMember
                self.RMA._derive_parameters(self.SCPCOA, self.Position, self.RadarCollection, self.ImageFormation)
            if self.Grid is not None:
                # noinspection PyProtectedMember
                self.Grid._derive_rma(self.RMA, self.GeoData, self.RadarCollection, self.ImageFormation, self.Position)

        self.define_geo_image_corners()
        self.define_geo_valid_data()
        if self.Radiometric is not None:
            # noinspection PyProtectedMember
            self.Radiometric._derive_parameters(self.Grid, self.SCPCOA)

    def get_transmit_band_name(self) -> str:
        """
        Gets the processed transmit band name.

        Returns
        -------
        str
        """

        if self.ImageFormation is None:
            return 'UN'
        return self.ImageFormation.get_transmit_band_name()

    def get_processed_polarization_abbreviation(self) -> str:
        """
        Gets the processed polarization abbreviation (two letters).

        Returns
        -------
        str
        """

        if self.ImageFormation is None:
            return 'UN'
        return self.ImageFormation.get_polarization_abbreviation()

    def get_processed_polarization(self) -> str:
        """
        Gets the processed polarization.

        Returns
        -------
        str
        """

        if self.ImageFormation is None:
            return 'UN'
        return self.ImageFormation.get_polarization()

    def apply_reference_frequency(self, reference_frequency: float) -> None:
        """
        If the reference frequency is used, adjust the necessary fields accordingly.

        Parameters
        ----------
        reference_frequency : float
            The reference frequency.

        Returns
        -------
        None
        """

        if self.RadarCollection is None:
            raise ValueError('RadarCollection is not defined. The reference frequency cannot be applied.')
        elif not self.RadarCollection.RefFreqIndex:  # it's None or 0
            raise ValueError(
                'RadarCollection.RefFreqIndex is not defined. The reference frequency should not be applied.')

        # noinspection PyProtectedMember
        self.RadarCollection._apply_reference_frequency(reference_frequency)
        if self.ImageFormation is not None:
            # noinspection PyProtectedMember
            self.ImageFormation._apply_reference_frequency(reference_frequency)
        if self.Antenna is not None:
            # noinspection PyProtectedMember
            self.Antenna._apply_reference_frequency(reference_frequency)
        if self.RMA is not None:
            # noinspection PyProtectedMember
            self.RMA._apply_reference_frequency(reference_frequency)

    def get_ground_resolution(self) -> Tuple[float, float]:
        """
        Gets the ground resolution for the sicd.

        Returns
        -------
        (float, float)
        """

        graze = numpy.deg2rad(self.SCPCOA.GrazeAng)
        twist = numpy.deg2rad(self.SCPCOA.TwistAng)
        row_ss = self.Grid.Row.SS
        col_ss = self.Grid.Col.SS

        row_ground = abs(float(row_ss / numpy.cos(graze)))
        col_ground = float(numpy.sqrt((numpy.tan(graze) * numpy.tan(twist) * row_ss)**2 + (col_ss/numpy.cos(twist))**2))
        return row_ground, col_ground

    def can_project_coordinates(self) -> bool:
        """
        Determines whether the necessary elements are populated to permit projection
        between image and physical coordinates. If False, then the (first discovered)
        reason why not will be logged at error level.

        Returns
        -------
        bool
        """

        if self._coa_projection is not None:
            return True

        # GeoData elements?
        if self.GeoData is None:
            logger.error('Formulating a projection is not feasible because GeoData is not populated.')
            return False
        if self.GeoData.SCP is None:
            logger.error('Formulating a projection is not feasible because GeoData.SCP is not populated.')
            return False
        if self.GeoData.SCP.ECF is None:
            logger.error('Formulating a projection is not feasible because GeoData.SCP.ECF is not populated.')
            return False

        # ImageData elements?
        if self.ImageData is None:
            logger.error('Formulating a projection is not feasible because ImageData is not populated.')
            return False
        if self.ImageData.FirstRow is None:
            logger.error('Formulating a projection is not feasible because ImageData.FirstRow is not populated.')
            return False
        if self.ImageData.FirstCol is None:
            logger.error('Formulating a projection is not feasible because ImageData.FirstCol is not populated.')
            return False
        if self.ImageData.SCPPixel is None:
            logger.error('Formulating a projection is not feasible because ImageData.SCPPixel is not populated.')
            return False
        if self.ImageData.SCPPixel.Row is None:
            logger.error('Formulating a projection is not feasible because ImageData.SCPPixel.Row is not populated.')
            return False
        if self.ImageData.SCPPixel.Col is None:
            logger.error('Formulating a projection is not feasible because ImageData.SCPPixel.Col is not populated.')
            return False

        # Position elements?
        if self.Position is None:
            logger.error('Formulating a projection is not feasible because Position is not populated.')
            return False
        if self.Position.ARPPoly is None:
            logger.error('Formulating a projection is not feasible because Position.ARPPoly is not populated.')
            return False

        # Grid elements?
        if self.Grid is None:
            logger.error('Formulating a projection is not feasible because Grid is not populated.')
            return False
        if self.Grid.TimeCOAPoly is None:
            logger.warning(
                'Formulating a projection may be inaccurate, because Grid.TimeCOAPoly is not populated and '
                'a constant approximation will be used.')
        if self.Grid.Row is None:
            logger.error('Formulating a projection is not feasible because Grid.Row is not populated.')
            return False
        if self.Grid.Row.SS is None:
            logger.error('Formulating a projection is not feasible because Grid.Row.SS is not populated.')
            return False
        if self.Grid.Col is None:
            logger.error('Formulating a projection is not feasible because Grid.Col is not populated.')
            return False
        if self.Grid.Col.SS is None:
            logger.error('Formulating a projection is not feasible because Grid.Col.SS is not populated.')
            return False
        if self.Grid.Type is None:
            logger.error('Formulating a projection is not feasible because Grid.Type is not populated.')
            return False

        # specifics for Grid.Type value
        if self.Grid.Type == 'RGAZIM':
            if self.ImageFormation is None:
                logger.error(
                    'Formulating a projection is not feasible because Grid.Type = "RGAZIM",\n\t'
                    'but ImageFormation is not populated.')
                return False
            if self.ImageFormation.ImageFormAlgo is None:
                logger.error(
                    'Formulating a projection is not feasible because Grid.Type = "RGAZIM",\n\t'
                    'but ImageFormation.ImageFormAlgo is not populated.')
                return False

            if self.ImageFormation.ImageFormAlgo == 'PFA':
                if self.PFA is None:
                    logger.error(
                        'ImageFormation.ImageFormAlgo is "PFA",\n\t'
                        'but the PFA parameter is not populated.\n\t'
                        'No projection can be done.')
                    return False
                if self.PFA.PolarAngPoly is None:
                    logger.error(
                        'ImageFormation.ImageFormAlgo is "PFA",\n\t'
                        'but the PFA.PolarAngPoly parameter is not populated.\n\t'
                        'No projection can be done.')
                    return False
                if self.PFA.SpatialFreqSFPoly is None:
                    logger.error(
                        'ImageFormation.ImageFormAlgo is "PFA",\n\t'
                        'but the PFA.SpatialFreqSFPoly parameter is not populated.\n\t'
                        'No projection can be done.')
                    return False
            elif self.ImageFormation.ImageFormAlgo == 'RGAZCOMP':
                if self.RgAzComp is None:
                    logger.error(
                        'ImageFormation.ImageFormAlgo is "RGAZCOMP",\n\t'
                        'but the RgAzComp parameter is not populated.\n\t'
                        'No projection can be done.')
                    return False
                if self.RgAzComp.AzSF is None:
                    logger.error(
                        'ImageFormation.ImageFormAlgo is "RGAZCOMP",\n\t'
                        'but the RgAzComp.AzSF parameter is not populated.\n\t'
                        'No projection can be done.')
                    return False
            else:
                logger.error(
                    'Grid.Type = "RGAZIM", and got unhandled ImageFormation.ImageFormAlgo {}.\n\t'
                    'No projection can be done.'.format(self.ImageFormation.ImageFormAlgo))
                return False
        elif self.Grid.Type == 'RGZERO':
            if self.RMA is None or self.RMA.INCA is None:
                logger.error(
                    'Grid.Type is "RGZERO", but the RMA.INCA parameter is not populated.\n\t'
                    'No projection can be done.')
                return False
            if self.RMA.INCA.R_CA_SCP is None or self.RMA.INCA.TimeCAPoly is None \
                    or self.RMA.INCA.DRateSFPoly is None:
                logger.error(
                    'Grid.Type is "RGZERO", but the parameters\n\t'
                    'R_CA_SCP, TimeCAPoly, or DRateSFPoly of RMA.INCA parameter are not populated.\n\t'
                    'No projection can be done.')
                return False
        elif self.Grid.Type in ['XRGYCR', 'XCTYAT', 'PLANE']:
            if self.Grid.Row.UVectECF is None or self.Grid.Col.UVectECF is None:
                logger.error(
                    'Grid.Type is one of ["XRGYCR", "XCTYAT", "PLANE"], but the UVectECF parameter of '
                    'Grid.Row or Grid.Col is not populated.\n\t'
                    'No projection can be formulated.')
                return False
        else:
            logger.error(
                'Unhandled Grid.Type {},\n\t'
                'unclear how to formulate a projection.'.format(self.Grid.Type))
            return False

        # logger.info('Consider calling sicd.define_coa_projection if the sicd structure is defined.')
        return True

    def define_coa_projection(
            self,
            delta_arp: Union[None, numpy.ndarray, list, tuple] = None,
            delta_varp: Union[None, numpy.ndarray, list, tuple] = None,
            range_bias: Optional[float] = None,
            adj_params_frame: str = 'ECF',
            override: bool = True) -> None:
        """
        Define the COAProjection object.

        Parameters
        ----------
        delta_arp : None|numpy.ndarray|list|tuple
            ARP position adjustable parameter (ECF, m).  Defaults to 0 in each coordinate.
        delta_varp : None|numpy.ndarray|list|tuple
            VARP position adjustable parameter (ECF, m/s).  Defaults to 0 in each coordinate.
        range_bias : float|int
            Range bias adjustable parameter (m), defaults to 0.
        adj_params_frame : str
            One of ['ECF', 'RIC_ECF', 'RIC_ECI'], specifying the coordinate frame used for
            expressing `delta_arp` and `delta_varp` parameters.
        override : bool
            should we redefine, if it is previously defined?

        Returns
        -------
        None
        """

        if not self.can_project_coordinates():
            logger.error('The COAProjection object cannot be defined.')
            return

        if self._coa_projection is not None and not override:
            return

        self._coa_projection = point_projection.COAProjection.from_sicd(
            self, delta_arp=delta_arp, delta_varp=delta_varp, range_bias=range_bias,
            adj_params_frame=adj_params_frame)

    def project_ground_to_image(
            self,
            coords: Union[numpy.ndarray, list, tuple],
            **kwargs) -> Tuple[numpy.ndarray, Union[numpy.ndarray, float], Union[numpy.ndarray, int]]:
        """
        Transforms a 3D ECF point to pixel (row/column) coordinates. This is
        implemented in accordance with the SICD Image Projections Description Document.
        **Really Scene-To-Image projection.**"

        Parameters
        ----------
        coords : numpy.ndarray|tuple|list
            ECF coordinate to map to scene coordinates, of size `N x 3`.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.ground_to_image` method.

        Returns
        -------
        image_points: numpy.ndarray
            The determined image point array, of size `N x 2`. Following the
            SICD convention, he upper-left pixel is [0, 0].
        delta_gpn: numpy.ndarray|float
            Residual ground plane displacement (m).
        iterations: numpy.ndarray|int
            The number of iterations performed.

        See Also
        --------
        sarpy.geometry.point_projection.ground_to_image
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.ground_to_image(coords, self, **kwargs)

    def project_ground_to_image_geo(
            self,
            coords: Union[numpy.ndarray, list, tuple],
            ordering: str = 'latlong',
            **kwargs) -> Tuple[numpy.ndarray, Union[numpy.ndarray, float], Union[numpy.ndarray, int]]:
        """
        Transforms a 3D Lat/Lon/HAE point to pixel (row/column) coordinates. This is
        implemented in accordance with the SICD Image Projections Description Document.
        **Really Scene-To-Image projection.**"

        Parameters
        ----------
        coords : numpy.ndarray|tuple|list
            ECF coordinate to map to scene coordinates, of size `N x 3`.
        ordering : str
            If 'longlat', then the input is `[longitude, latitude, hae]`.
            Otherwise, the input is `[latitude, longitude, hae]`. Passed through
            to :func:`sarpy.geometry.geocoords.geodetic_to_ecf`.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.ground_to_image_geo` method.

        Returns
        -------
        image_points: numpy.ndarray
            The determined image point array, of size `N x 2`. Following the
            SICD convention, he upper-left pixel is [0, 0].
        delta_gpn: numpy.ndarray|float
            Residual ground plane displacement (m).
        iterations: numpy.ndarray|int
            The number of iterations performed.

        See Also
        --------
        sarpy.geometry.point_projection.ground_to_image_geo
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.ground_to_image_geo(coords, self, ordering=ordering, **kwargs)

    def project_image_to_ground(
            self,
            im_points: Union[numpy.ndarray, list, tuple],
            projection_type: str = 'HAE',
            **kwargs) -> numpy.ndarray:
        """
        Transforms image coordinates to ground plane ECF coordinate via the algorithm(s)
        described in SICD Image Projections document.

        Parameters
        ----------
        im_points : numpy.ndarray|list|tuple
            the image coordinate array
        projection_type : str
            One of `['PLANE', 'HAE', 'DEM']`. Type `DEM` is a work in progress.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.image_to_ground` method.

        Returns
        -------
        numpy.ndarray
            Ground Plane Point (in ECF coordinates) corresponding to the input image coordinates.

        See Also
        --------
        sarpy.geometry.point_projection.image_to_ground
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.image_to_ground(
            im_points, self, projection_type=projection_type, **kwargs)

    def project_image_to_ground_geo(
            self,
            im_points: Union[numpy.ndarray, list, tuple],
            ordering: str = 'latlong',
            projection_type: str = 'HAE',
            **kwargs) -> numpy.ndarray:
        """
        Transforms image coordinates to ground plane WGS-84 coordinate via the algorithm(s)
        described in SICD Image Projections document.

        Parameters
        ----------
        im_points : numpy.ndarray|list|tuple
            the image coordinate array
        projection_type : str
            One of `['PLANE', 'HAE', 'DEM']`. Type `DEM` is a work in progress.
        ordering : str
            Determines whether return is ordered as `[lat, long, hae]` or `[long, lat, hae]`.
            Passed through to :func:`sarpy.geometry.geocoords.ecf_to_geodetic`.
        kwargs
            The keyword arguments for the :func:`sarpy.geometry.point_projection.image_to_ground_geo` method.

        Returns
        -------
        numpy.ndarray
            Ground Plane Point (in ECF coordinates) corresponding to the input image coordinates.
        See Also
        --------
        sarpy.geometry.point_projection.image_to_ground_geo
        """

        if 'use_structure_coa' not in kwargs:
            kwargs['use_structure_coa'] = True
        return point_projection.image_to_ground_geo(
            im_points, self, ordering=ordering, projection_type=projection_type, **kwargs)

    def populate_rniirs(
            self,
            signal: Optional[float] = None,
            noise: Optional[float] = None,
            override: bool = False) -> None:
        """
        Given the signal and noise values (in sigma zero power units),
        calculate and populate an estimated RNIIRS value.

        Parameters
        ----------
        signal : None|float
        noise : None|float
        override : bool
            Override the value, if present.

        Returns
        -------
        None
        """

        from sarpy.processing.sicd.rgiqe import populate_rniirs_for_sicd
        populate_rniirs_for_sicd(self, signal=signal, noise=noise, override=override)

    def get_suggested_name(
            self,
            product_number: int = 1) -> str:
        """
        Get the suggested name stem for the sicd and derived data.

        Parameters
        ----------
        product_number : int

        Returns
        -------
        str
        """

        sugg_name = get_sicd_name(self, product_number)
        if sugg_name is None:
            sugg_name = self.CollectionInfo.CoreName if self.CollectionInfo.CoreName is not None else \
                'Unknown_Sicd{}'.format(product_number)
        return re.sub(':', '_', sugg_name)

    def version_required(self) -> Tuple[int, int, int]:
        """
        What SICD version is required for valid support?

        Returns
        -------
        tuple
        """

        required = (1, 1, 0)
        for fld in self._fields:
            val = getattr(self, fld)
            if val is not None and hasattr(val, 'version_required'):
                required = max(required, val.version_required())
        return required

    def get_des_details(
            self,
            check_older_version: bool = False) -> Dict:
        """
        Gets the correct current SICD DES subheader details.

        Parameters
        ----------
        check_older_version : bool
            If True and compatible, then version 1.1.0 information will be returned.
            Otherwise, the most recent supported version will be returned.

        Returns
        -------
        dict
        """

        required_version = self.version_required()
        # noinspection PyTypeChecker
        if required_version > _SICD_DEFAULT_TUPLE or check_older_version:
            info = _SICD_SPEC_DETAILS['{}.{}.{}'.format(*required_version)]
        else:
            info = _SICD_SPEC_DETAILS[_SICD_VERSION_DEFAULT]
        spec_ns = info['namespace']
        details = info['details']
        spec_version = details['version']
        spec_date = details['date']

        return OrderedDict([
            ('DESSHSI', _SICD_SPECIFICATION_IDENTIFIER),
            ('DESSHSV', spec_version),
            ('DESSHSD', spec_date),
            ('DESSHTN', spec_ns)])

    def copy(self):
        """
        Provides a deep copy.

        Returns
        -------
        SICDType
        """

        out = super(SICDType, self).copy()
        out._NITF = deepcopy(self._NITF)
        return out

    def to_xml_bytes(self, urn=None, tag='SICD', check_validity=False, strict=DEFAULT_STRICT):
        if urn is None:
            urn = _SICD_SPEC_DETAILS[_SICD_VERSION_DEFAULT]['namespace']
        return super(SICDType, self).to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict)

    def to_xml_string(self, urn=None, tag='SICD', check_validity=False, strict=DEFAULT_STRICT):
        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')

    def create_subset_structure(
            self,
            row_bounds: Optional[Tuple[int, int]] = None,
            column_bounds: Optional[Tuple[int, int]] = None):
        """
        Create a version of the SICD structure for a given subset.

        Parameters
        ----------
        row_bounds : None|tuple
        column_bounds : None|tuple

        Returns
        -------
        sicd : SICDType
            The sicd
        row_bounds : tuple
            Vetted tuple of the form `(min row, max row)`.
        column_bounds : tuple
            Vetted tuple of the form `(min column, max column)`.
        """

        sicd = self.copy()
        num_rows = self.ImageData.NumRows
        num_cols = self.ImageData.NumCols
        if row_bounds is not None:
            start_row = int(row_bounds[0])
            end_row = int(row_bounds[1])
            if not (0 <= start_row < end_row <= num_rows):
                raise ValueError(
                    'row bounds ({}, {}) are not sensible for NumRows {}'.format(
                        start_row, end_row, num_rows))
            sicd.ImageData.FirstRow = sicd.ImageData.FirstRow + start_row
            sicd.ImageData.NumRows = (end_row - start_row)
            out_row_bounds = (start_row, end_row)
        else:
            out_row_bounds = (0, num_rows)

        if column_bounds is not None:
            start_col = int(column_bounds[0])
            end_col = int(column_bounds[1])
            if not (0 <= start_col < end_col <= num_cols):
                raise ValueError(
                    'column bounds ({}, {}) are not sensible for NumCols {}'.format(
                        start_col, end_col, num_cols))
            sicd.ImageData.FirstCol = sicd.ImageData.FirstCol + start_col
            sicd.ImageData.NumCols = (end_col - start_col)
            out_col_bounds = (start_col, end_col)
        else:
            out_col_bounds = (0, num_cols)
        sicd.define_geo_image_corners(override=True)
        return sicd, out_row_bounds, out_col_bounds

    @classmethod
    def from_xml_file(cls, file_path):
        """
        Construct the sicd object from a stand-alone xml file path.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        SICDType
        """

        root_node, xml_ns = parse_xml_from_file(file_path)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Construct the sicd object from a xml string.

        Parameters
        ----------
        xml_string : str|bytes

        Returns
        -------
        SICDType
        """

        root_node, xml_ns = parse_xml_from_string(xml_string)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)



"""
The Per Vector parameters (PVP) definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List, Tuple, Optional

import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringDescriptor, IntegerDescriptor, \
    SerializableDescriptor, SerializableListDescriptor
from .utils import binary_format_string_to_dtype, homogeneous_dtype
from .base import DEFAULT_STRICT


class PerVectorParameterI8(Serializable):
    _fields = ('Offset', 'Size', 'Format')
    _required = _fields
    # descriptors
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The offset value.')  # type: int

    def __init__(self, Offset=None, **kwargs):
        """

        Parameters
        ----------
        Offset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Offset = Offset
        super(PerVectorParameterI8, self).__init__(**kwargs)

    @property
    def Size(self):
        """
        int: The size of the vector, constant value 1 here.
        """

        return 1

    @property
    def Format(self):
        """
        str: The format of the vector data, constant value 'I8' here.
        """

        return 'I8'


class PerVectorParameterF8(Serializable):
    _fields = ('Offset', 'Size', 'Format')
    _required = _fields
    # descriptors
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The offset value.')  # type: int

    def __init__(self, Offset=None, **kwargs):
        """

        Parameters
        ----------
        Offset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Offset = Offset
        super(PerVectorParameterF8, self).__init__(**kwargs)

    @property
    def Size(self):
        """
        int: The size of the vector, constant value 1 here.
        """

        return 1

    @property
    def Format(self):
        """
        str: The format of the vector data, constant value 'F8' here.
        """

        return 'F8'


class PerVectorParameterXYZ(Serializable):
    _fields = ('Offset', 'Size', 'Format')
    _required = _fields
    # descriptors
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The offset value.')  # type: int

    def __init__(self, Offset=None, **kwargs):
        """

        Parameters
        ----------
        Offset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Offset = Offset
        super(PerVectorParameterXYZ, self).__init__(**kwargs)

    @property
    def Size(self):
        """
        int: The size of the vector, constant value 3 here.
        """

        return 3

    @property
    def Format(self):
        """
        str: The format of the vector data, constant value 'X=F8;Y=F8;Z=F8;' here.
        """

        return 'X=F8;Y=F8;Z=F8;'


class PerVectorParameterEB(Serializable):
    _fields = ('Offset', 'Size', 'Format')
    _required = _fields
    # descriptors
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The offset value.')  # type: int

    def __init__(self, Offset=None, **kwargs):
        """

        Parameters
        ----------
        Offset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Offset = Offset
        super(PerVectorParameterEB, self).__init__(**kwargs)

    @property
    def Size(self):
        """
        int: The size of the vector, constant value 2 here.
        """

        return 2

    @property
    def Format(self):
        """
        str: The format of the vector data, constant value 'DCX=F8;DCY=F8;' here.
        """

        return 'DCX=F8;DCY=F8;'


class UserDefinedPVPType(Serializable):
    """
    A user defined PVP structure.
    """

    _fields = ('Name', 'Offset', 'Size', 'Format')
    _required = _fields
    # descriptors
    Name = StringDescriptor(
        'Name', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='')  # type: int
    Size = IntegerDescriptor(
        'Size', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='')  # type: int
    Format = StringDescriptor(
        'Format', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str

    def __init__(self, Name=None, Offset=None, Size=None, Format=None, **kwargs):
        """

        Parameters
        ----------
        Name : str
        Offset : int
        Size : int
        Format : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.Offset = Offset
        self.Size = Size
        self.Format = Format
        super(UserDefinedPVPType, self).__init__(**kwargs)


class TxAntennaType(Serializable):
    _fields = ('TxACX', 'TxACY', 'TxEB')
    _required = _fields
    TxACX = SerializableDescriptor(
        'TxACX', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    TxACY = SerializableDescriptor(
        'TxACY', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    TxEB = SerializableDescriptor(
        'TxEB', PerVectorParameterEB, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterEB

    def __init__(
            self,
            TxACX: PerVectorParameterXYZ = None,
            TxACY: PerVectorParameterXYZ = None,
            TxEB: PerVectorParameterEB = None,
            **kwargs):
        """
        Parameters
        ----------
        TxACX : PerVectorParameterXYZ
        TxACY : PerVectorParameterXYZ
        TxEB : PerVectorParameterEB
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxACX = TxACX
        self.TxACY = TxACY
        self.TxEB = TxEB
        super(TxAntennaType, self).__init__(**kwargs)


class RcvAntennaType(Serializable):
    _fields = ('RcvACX', 'RcvACY', 'RcvEB')
    _required = _fields
    RcvACX = SerializableDescriptor(
        'RcvACX', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvACY = SerializableDescriptor(
        'RcvACY', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvEB = SerializableDescriptor(
        'RcvEB', PerVectorParameterEB, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterEB

    def __init__(
            self,
            RcvACX: Optional[PerVectorParameterXYZ] = None,
            RcvACY: Optional[PerVectorParameterXYZ] = None,
            RcvEB: Optional[PerVectorParameterEB] = None,
            **kwargs):
        """
        Parameters
        ----------
        RcvACX : PerVectorParameterXYZ
        RcvACY : PerVectorParameterXYZ
        RcvEB : PerVectorParameterEB
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvACX = RcvACX
        self.RcvACY = RcvACY
        self.RcvEB = RcvEB
        super(RcvAntennaType, self).__init__(**kwargs)


class PVPType(Serializable):
    _fields = (
        'TxTime', 'TxPos', 'TxVel', 'RcvTime', 'RcvPos', 'RcvVel',
        'SRPPos', 'AmpSF', 'aFDOP', 'aFRR1', 'aFRR2', 'FX1', 'FX2',
        'FXN1', 'FXN2', 'TOA1', 'TOA2', 'TOAE1', 'TOAE2', 'TDTropoSRP',
        'TDIonoSRP', 'SC0', 'SCSS', 'SIGNAL',
        'TxAntenna', 'RcvAntenna', 'AddedPVP')
    _required = (
        'TxTime', 'TxPos', 'TxVel', 'RcvTime', 'RcvPos', 'RcvVel',
        'SRPPos', 'aFDOP', 'aFRR1', 'aFRR2', 'FX1', 'FX2',
        'TOA1', 'TOA2', 'TDTropoSRP', 'SC0', 'SCSS')
    _collections_tags = {'AddedPVP': {'array': False, 'child_tag': 'AddedPVP'}}
    # descriptors
    TxTime = SerializableDescriptor(
        'TxTime', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TxPos = SerializableDescriptor(
        'TxPos', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    TxVel = SerializableDescriptor(
        'TxVel', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvTime = SerializableDescriptor(
        'RcvTime', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    RcvPos = SerializableDescriptor(
        'RcvPos', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvVel = SerializableDescriptor(
        'RcvVel', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    SRPPos = SerializableDescriptor(
        'SRPPos', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    AmpSF = SerializableDescriptor(
        'AmpSF', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    aFDOP = SerializableDescriptor(
        'aFDOP', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    aFRR1 = SerializableDescriptor(
        'aFRR1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    aFRR2 = SerializableDescriptor(
        'aFRR2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FX1 = SerializableDescriptor(
        'FX1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FX2 = SerializableDescriptor(
        'FX2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FXN1 = SerializableDescriptor(
        'FXN1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FXN2 = SerializableDescriptor(
        'FXN2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TOA1 = SerializableDescriptor(
        'TOA1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TOA2 = SerializableDescriptor(
        'TOA2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TOAE1 = SerializableDescriptor(
        'TOAE1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TOAE2 = SerializableDescriptor(
        'TOAE2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TDTropoSRP = SerializableDescriptor(
        'TDTropoSRP', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TDIonoSRP = SerializableDescriptor(
        'TDIonoSRP', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    SC0 = SerializableDescriptor(
        'SC0', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    SCSS = SerializableDescriptor(
        'SCSS', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    SIGNAL = SerializableDescriptor(
        'SIGNAL', PerVectorParameterI8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterI8
    TxAntenna = SerializableDescriptor(
        'TxAntenna', TxAntennaType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[TxAntennaType]
    RcvAntenna = SerializableDescriptor(
        'RcvAntenna', RcvAntennaType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[RcvAntennaType]
    AddedPVP = SerializableListDescriptor(
        'AddedPVP', UserDefinedPVPType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, List[UserDefinedPVPType]]

    def __init__(
            self,
            TxTime: PerVectorParameterF8 = None,
            TxPos: PerVectorParameterXYZ = None,
            TxVel: PerVectorParameterXYZ = None,
            RcvTime: PerVectorParameterF8 = None,
            RcvPos: PerVectorParameterXYZ = None,
            RcvVel: PerVectorParameterXYZ = None,
            SRPPos: PerVectorParameterXYZ = None,
            AmpSF: Optional[PerVectorParameterF8] = None,
            aFDOP: PerVectorParameterF8 = None,
            aFRR1: PerVectorParameterF8 = None,
            aFRR2: PerVectorParameterF8 = None,
            FX1: PerVectorParameterF8 = None,
            FX2: PerVectorParameterF8 = None,
            FXN1: Optional[PerVectorParameterF8] = None,
            FXN2: Optional[PerVectorParameterF8] = None,
            TOA1: PerVectorParameterF8 = None,
            TOA2: PerVectorParameterF8 = None,
            TOAE1: Optional[PerVectorParameterF8] = None,
            TOAE2: Optional[PerVectorParameterF8] = None,
            TDTropoSRP: PerVectorParameterF8 = None,
            TDIonoSRP: Optional[PerVectorParameterF8] = None,
            SC0: PerVectorParameterF8 = None,
            SCSS: PerVectorParameterF8 = None,
            SIGNAL: Optional[PerVectorParameterI8] = None,
            TxAntenna: Optional[TxAntennaType] = None,
            RcvAntenna: Optional[RcvAntennaType] = None,
            AddedPVP: Optional[List[UserDefinedPVPType]] = None,
            **kwargs):
        """

        Parameters
        ----------
        TxTime : PerVectorParameterF8
        TxPos : PerVectorParameterXYZ
        TxVel : PerVectorParameterXYZ
        RcvTime : PerVectorParameterF8
        RcvPos : PerVectorParameterXYZ
        RcvVel : PerVectorParameterXYZ
        SRPPos : PerVectorParameterXYZ
        AmpSF : None|PerVectorParameterF8
        aFDOP : PerVectorParameterF8
        aFRR1 : PerVectorParameterF8
        aFRR2 : PerVectorParameterF8
        FX1 : PerVectorParameterF8
        FX2 : PerVectorParameterF8
        FXN1 : None|PerVectorParameterF8
        FXN2 : None|PerVectorParameterF8
        TOA1 : PerVectorParameterF8
        TOA2 : PerVectorParameterF8
        TOAE1 : None|PerVectorParameterF8
        TOAE2 : None|PerVectorParameterF8
        TDTropoSRP : PerVectorParameterF8
        TDIonoSRP : None|PerVectorParameterF8
        SC0 : PerVectorParameterF8
        SCSS : PerVectorParameterF8
        SIGNAL : None|PerVectorParameterI8
        TxAntenna : None|TxAntennaType
        RcvAntenna : None|RcvAntennaType
        AddedPVP : None|List[UserDefinedPVPType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxTime = TxTime
        self.TxPos = TxPos
        self.TxVel = TxVel
        self.RcvTime = RcvTime
        self.RcvPos = RcvPos
        self.RcvVel = RcvVel
        self.SRPPos = SRPPos
        self.AmpSF = AmpSF
        self.aFDOP = aFDOP
        self.aFRR1 = aFRR1
        self.aFRR2 = aFRR2
        self.FX1 = FX1
        self.FX2 = FX2
        self.FXN1 = FXN1
        self.FXN2 = FXN2
        self.TOA1 = TOA1
        self.TOA2 = TOA2
        self.TOAE1 = TOAE1
        self.TOAE2 = TOAE2
        self.TDTropoSRP = TDTropoSRP
        self.TDIonoSRP = TDIonoSRP
        self.SC0 = SC0
        self.SCSS = SCSS
        self.SIGNAL = SIGNAL
        self.TxAntenna = TxAntenna
        self.RcvAntenna = RcvAntenna
        self.AddedPVP = AddedPVP
        super(PVPType, self).__init__(**kwargs)

    def get_size(self):
        """
        Gets the size in bytes of each vector.

        Returns
        -------
        int
        """

        out = 0
        for fld in self._fields[:-3]:
            val = getattr(self, fld)
            if val is not None:
                out += val.Size*8
        for fld in ['TxAntenna', 'RcvAntenna']:
            val = getattr(self, fld)
            if val is not None:
                assert isinstance(val, (TxAntennaType, RcvAntennaType))
                out += (3 + 3 + 2)*8
        if self.AddedPVP is not None:
            for entry in self.AddedPVP:
                out += entry.Size*8
        return out

    def get_offset_size_format(self, field):
        """
        Get the Offset (in bytes), Size (in bytes) for the given field,
        as well as the corresponding struct format string.

        Parameters
        ----------
        field : str
            The desired field name.

        Returns
        -------
        None|Tuple[int, int, str]
        """

        def get_return(the_val) -> Union[None, Tuple[int, int, str]]:
            if the_val is None:
                return None
            return the_val.Offset*8, the_val.Size*8, homogeneous_dtype(the_val.Format).char

        if field in self._fields[:-3]:
            return get_return(getattr(self, field))
        elif field in ['TxACX', 'TxACY', 'TxEB']:
            if self.TxAntenna is None:
                return None
            else:
                return get_return(getattr(self.TxAntenna, field))
        elif field in ['RcvACX', 'RcvACY', 'RcvEB']:
            if self.RcvAntenna is None:
                return None
            else:
                return get_return(getattr(self.RcvAntenna, field))
        else:
            if self.AddedPVP is None:
                return None
            for val in self.AddedPVP:
                if field == val.Name:
                    return get_return(val)
            return None

    def get_vector_dtype(self):
        """
        Gets the dtype for the corresponding structured array for the full PVP
        array.

        Returns
        -------
        numpy.dtype
            This will be a compound dtype for a structured array.
        """

        bytes_per_word = 8
        names = []
        formats = []
        offsets = []

        for fld in self._fields:
            val = getattr(self, fld)
            if val is None:
                continue
            elif fld == 'TxAntenna':
                for t_fld in ['TxACX', 'TxACY', 'TxEB']:
                    t_val = getattr(val, t_fld)
                    names.append(t_fld)
                    formats.append(binary_format_string_to_dtype(t_val.Format))
                    offsets.append(t_val.Offset*bytes_per_word)
            elif fld == 'RcvAntenna':
                for t_fld in ['RcvACX', 'RcvACY', 'RcvEB']:
                    t_val = getattr(val, t_fld)
                    names.append(t_fld)
                    formats.append(binary_format_string_to_dtype(t_val.Format))
                    offsets.append(t_val.Offset*bytes_per_word)
            elif fld == 'AddedPVP':
                for entry in val:
                    assert isinstance(entry, UserDefinedPVPType)
                    names.append(entry.Name)
                    formats.append(binary_format_string_to_dtype(entry.Format))
                    offsets.append(entry.Offset*bytes_per_word)
            else:
                names.append(fld)
                formats.append(binary_format_string_to_dtype(val.Format))
                offsets.append(val.Offset*bytes_per_word)
        return numpy.dtype({'names': names, 'formats': formats, 'offsets': offsets})

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.TxAntenna is not None or self.RcvAntenna is not None:
            required = max(required, (1, 1, 0))
        return required
